#include "AStarPlanner.h"

#include <algorithm>
#include <cmath>

#include <boost/geometry.hpp>
#include <boost/geometry/algorithms/detail/intersects/interface.hpp>

#include <Eigen/Geometry>

#include <IceUtil/Time.h>

#include <VirtualRobot/BoundingBox.h>
#include <VirtualRobot/CollisionDetection/CollisionChecker.h>
#include <VirtualRobot/CollisionDetection/CollisionModel.h>
#include <VirtualRobot/XML/ObjectIO.h>

#include "ArmarXCore/core/exceptions/local/ExpressionException.h"
#include "ArmarXCore/core/logging/Logging.h"

#include "Navigation/libraries/conversions/eigen.h"

namespace bg = boost::geometry;

namespace armarx::nav::algo::astar
{

    bool intersects(const VirtualRobot::BoundingBox& bb1, const VirtualRobot::BoundingBox& bb2)
    {
        using point_t = bg::model::point<double, 3, bg::cs::cartesian>;
        using box_t = bg::model::box<point_t>;

        const auto toPoint = [](const Eigen::Vector3f& pt)
        { return point_t(pt.x(), pt.y(), pt.z()); };

        box_t box1(toPoint(bb1.getMin()), toPoint(bb1.getMax()));
        box_t box2(toPoint(bb2.getMin()), toPoint(bb2.getMax()));

        return bg::intersects(box1, box2);
    }

    AStarPlanner::AStarPlanner(VirtualRobot::RobotPtr robot,
                               VirtualRobot::SceneObjectSetPtr obstacles,
                               float cellSize) :
        Planner2D(robot, obstacles), cellSize(cellSize)
    {
    }

    float AStarPlanner::heuristic(NodePtr n1, NodePtr n2)
    {
        return (n1->position - n2->position).norm();
    }

    void AStarPlanner::createUniformGrid()
    {
        ARMARX_TRACE;

        ARMARX_INFO << "Scene bounds are " << sceneBoundsMin << " and " << sceneBoundsMax;

        //+1 for explicit rounding up
        cols = (sceneBoundsMax.x() - sceneBoundsMin.x()) / cellSize + 1;
        rows = (sceneBoundsMax.y() - sceneBoundsMin.y()) / cellSize + 1;

        ARMARX_INFO << "Grid size: " << rows << ", " << cols;

        for (size_t r = 0; r < rows; r++)
        {
            grid.push_back(std::vector<NodePtr>(cols));
            for (size_t c = 0; c < cols; c++)
            {
                Eigen::Vector2f pos;
                pos.x() = sceneBoundsMin.x() + c * cellSize + cellSize / 2;
                pos.y() = sceneBoundsMin.y() + r * cellSize + cellSize / 2;
                grid[r][c] = NodePtr(new Node(pos));
            }
        }

        // Init successors
        for (size_t r = 0; r < rows; r++)
        {
            for (size_t c = 0; c < cols; c++)
            {
                NodePtr candidate = grid[r][c];
                if (!fulfillsConstraints(candidate))
                {
                    continue;
                }

                // Add the valid node as successor to all its neighbors
                for (int nR = -1; nR <= 1; nR++)
                {
                    for (int nC = -1; nC <= 1; nC++)
                    {
                        const int neighborIndexX = c + nC;
                        const int neighborIndexY = r + nR;
                        if (neighborIndexX < 0 || neighborIndexY < 0 || (nR == 0 && nC == 0))
                        {
                            continue;
                        }

                        if (neighborIndexY >= static_cast<int>(rows) ||
                            neighborIndexX >= static_cast<int>(cols))
                        {
                            continue;
                        }

                        grid[neighborIndexY][neighborIndexX]->successors.push_back(candidate);
                    }
                }
            }
        }
    }

    bool AStarPlanner::fulfillsConstraints(NodePtr n)
    {
        ARMARX_TRACE;
        ARMARX_CHECK_NOT_NULL(n);
        ARMARX_CHECK_NOT_NULL(robotCollisionModel);
        ARMARX_CHECK_NOT_NULL(VirtualRobot::CollisionChecker::getGlobalCollisionChecker());
        ARMARX_CHECK_NOT_NULL(obstacles);

        const Eigen::Affine3f globalPose =
            Eigen::Translation3f(conv::to3D(n->position)) *
            Eigen::AngleAxisf(M_PI_2f32, Eigen::Vector3f::UnitX());

        robotCollisionModel->setGlobalPose(globalPose.matrix());

        VirtualRobot::BoundingBox robotBbox = robotCollisionModel->getBoundingBox(true);

        Eigen::Vector3f P1, P2;
        int id1, id2;
        for (size_t i = 0; i < obstacles->getSize(); i++)
        {
            // cheap collision check
            VirtualRobot::BoundingBox obstacleBbox =
                obstacles->getSceneObject(i)->getCollisionModel()->getBoundingBox(true);
            if (not intersects(robotBbox, obstacleBbox))
            {
                continue;
            }

            // precise collision check
            float dist =
                VirtualRobot::CollisionChecker::getGlobalCollisionChecker()->calculateDistance(
                    robotCollisionModel,
                    obstacles->getSceneObject(i)->getCollisionModel(),
                    P1,
                    P2,
                    &id1,
                    &id2);

            if (dist <= cellSize / 2)
            {
                return false;
            }
        }
        return n->position.x() >= sceneBoundsMin.x() && n->position.x() <= sceneBoundsMax.x() &&
               n->position.y() >= sceneBoundsMin.y() && n->position.y() <= sceneBoundsMax.y();
    }

    NodePtr AStarPlanner::closestNode(Eigen::Vector2f v)
    {
        float r = (v.y() - cellSize / 2 - sceneBoundsMin.y()) / cellSize;
        float c = (v.x() - cellSize / 2 - sceneBoundsMin.x()) / cellSize;

        ARMARX_CHECK(r >= 0.F);
        ARMARX_CHECK(c >= 0.F);

        ARMARX_CHECK(r <= (rows - 1));
        ARMARX_CHECK(c <= (cols - 1));

        return grid[static_cast<int>(r)][static_cast<int>(c)];
    }

    std::vector<Eigen::Vector2f> AStarPlanner::plan(Eigen::Vector2f start, Eigen::Vector2f goal)
    {
        ARMARX_TRACE;
        grid.clear();
        std::vector<Eigen::Vector2f> result;

        ARMARX_CHECK_NOT_NULL(robot);

        // ARMARX_CHECK(robot->hasRobotNode(robotColModelName)) << "No robot node " << robotColModelName;
        // ARMARX_CHECK(robot->getRobotNode(robotColModelName)->getCollisionModel())
        // << "No collision model for robot node " << robotColModelName;

        // robotCollisionModel = robot->getRobotNode(robotColModelName)->getCollisionModel()->clone();
        robotCollisionModel->setGlobalPose(robot->getGlobalPose());

        // static auto cylinder = VirtualRobot::ObjectIO::loadObstacle("/home/fabi/cylinder.xml");
        // ARMARX_CHECK_NOT_NULL(cylinder);
        // cylinder->setGlobalPose(robot->getGlobalPose());

        // robotCollisionModel = cylinder->getCollisionModel();

        ARMARX_CHECK_NOT_NULL(robotCollisionModel);

        createUniformGrid();
        ARMARX_INFO << "Setting start node";
        NodePtr nodeStart = closestNode(start);

        ARMARX_INFO << "Setting goal node";
        NodePtr nodeGoal = closestNode(goal);

        std::vector<NodePtr> closedSet;
        std::vector<NodePtr> openSet;
        openSet.push_back(nodeStart);
        std::map<NodePtr, float> gScore;
        gScore[nodeStart] = 0;
        std::map<NodePtr, float> fScore;
        fScore[nodeStart] = gScore[nodeStart] + heuristic(nodeStart, nodeGoal);

        while (!openSet.empty())
        {
            float lowestScore = fScore[openSet[0]];
            auto currentIT = openSet.begin();
            for (auto it = openSet.begin() + 1; it != openSet.end(); it++)
            {
                if (fScore[*it] < lowestScore)
                {
                    lowestScore = fScore[*it];
                    currentIT = it;
                }
            }
            NodePtr currentBest = *currentIT;
            if (currentBest == nodeGoal)
            {
                break;
            }

            openSet.erase(currentIT);
            closedSet.push_back(currentBest);
            for (size_t i = 0; i < currentBest->successors.size(); i++)
            {
                NodePtr neighbor = currentBest->successors[i];
                if (std::find(closedSet.begin(), closedSet.end(), neighbor) != closedSet.end())
                {
                    continue;
                }

                float tentativeGScore = gScore[currentBest] + heuristic(currentBest, neighbor);
                bool notInOS = std::find(openSet.begin(), openSet.end(), neighbor) == openSet.end();
                if (notInOS || tentativeGScore < gScore[neighbor])
                {
                    neighbor->predecessor = currentBest;
                    gScore[neighbor] = tentativeGScore;
                    fScore[neighbor] = tentativeGScore + heuristic(neighbor, nodeGoal);
                    if (notInOS)
                    {
                        openSet.push_back(neighbor);
                    }
                }
            }
        }

        // Found solution, now retrieve path from goal to start
        if (nodeGoal)
        {
            result = nodeGoal->traversePredecessors();
        }

        // Since the graph was traversed from goal to start, we have to reverse the order
        std::reverse(result.begin(), result.end());

        return result;
    }

    void AStarPlanner::setRobotCollisionModel(const VirtualRobot::CollisionModelPtr& collisionModel)
    {
        robotCollisionModel = collisionModel;
    }

} // namespace armarx::nav::algo::astar
